package de.ozgcloud.nachrichten.antragraum;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.opensaml.saml.saml2.core.Response;

import de.ozgcloud.nachrichten.NachrichtenManagerProperties;
import de.ozgcloud.nachrichten.common.vorgang.VorgangService;
import de.ozgcloud.nachrichten.common.vorgang.VorgangTestFactory;
import de.ozgcloud.nachrichten.postfach.PersistPostfachNachrichtService;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;

class AntragraumServiceTest {
	@Spy
	@InjectMocks
	private AntragraumService service;

	@Mock
	private AntragraumProperties properties;
	@Mock
	private NachrichtenManagerProperties nachrichtenManagerProperties;
	@Mock
	private PersistPostfachNachrichtService postfachNachrichtService;
	@Mock
	private Saml2Verifier verifier;
	@Mock
	private Saml2Parser parser;
	@Mock
	private Saml2Decrypter decrypter;
	@Mock
	private RueckfrageMapper rueckfrageMapper;
	@Mock
	private VorgangService vorgangService;

	@Nested
	class TestGetAntragsraumUrl {

		private static final String URL = "url";

		@Test
		void shouldReturnUrl() {
			when(properties.getUrl()).thenReturn(URL);

			var url = service.getAntragsraumUrl();

			assertThat(url).isEqualTo(URL);
		}
	}

	@DisplayName("Find rueckfragen")
	@Nested
	class TestFindRueckfragen {

		@Mock
		private Response response;

		private static final String SAML_TOKEN = "TOKEN";
		private static final String POSTFACH_ID = "dummyPostfachId";

		private final Rueckfrage rueckfrage = RueckfrageTestFactory.create();
		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void mock() {
			when(postfachNachrichtService.findRueckfragen(any(), any())).thenReturn(Stream.of(postfachNachricht));

			when(rueckfrageMapper.fromPostfachNachricht(any())).thenReturn(rueckfrage);
			doReturn(rueckfrage).when(service).enrichByAnswers(any(), any());
			doReturn(Stream.of(rueckfrage)).when(service).enrichRueckfragen(any());
			doReturn(POSTFACH_ID).when(service).getPostfachId(any());
		}

		@Test
		void shouldCallVerify() {
			service.findRueckfragen(SAML_TOKEN).toList();

			verify(service).verifyToken(SAML_TOKEN);
		}

		@Test
		void shouldCallVerifier() {
			service.findRueckfragen(SAML_TOKEN).toList();

			verify(verifier).verify(anyString());
		}

		@Test
		void shouldGetPostfachId() {
			service.findRueckfragen(SAML_TOKEN).toList();

			verify(service).getPostfachId(SAML_TOKEN);
		}

		@Test
		void shouldCallPostfachService() {
			service.findRueckfragen(SAML_TOKEN).toList();

			verify(postfachNachrichtService).findRueckfragen(AntragraumService.BAYERN_ID_SERVICE_KONTO_TYPE, POSTFACH_ID);
		}

		@Test
		void shouldCallMapper() {
			service.findRueckfragen(SAML_TOKEN).toList();

			verify(rueckfrageMapper).fromPostfachNachricht(postfachNachricht);
		}

		@Test
		void shouldCallEnrichByAnswers() {
			service.findRueckfragen(SAML_TOKEN).toList();

			verify(service).enrichByAnswers(rueckfrage, SAML_TOKEN);
		}

		@Test
		void shouldCallEnrichRueckfragen() {
			service.findRueckfragen(SAML_TOKEN).toList();

			verify(service).enrichRueckfragen(any());
		}

		@Test
		void shouldReturnFullFilledValue() {
			var rueckfragen = service.findRueckfragen(SAML_TOKEN).toList();

			assertThat(rueckfragen).containsExactly(rueckfrage);
		}
	}

	@DisplayName("Enrich by answers")
	@Nested
	class TestEnrichByAnswers {

		private static final String SAML_TOKEN = "TOKEN";
		private final Rueckfrage rueckfrage = RueckfrageTestFactory.create();
		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();
		private final Stream<PostfachNachricht> answers = Stream.of(postfachNachricht);

		@BeforeEach
		void mock() {
			doReturn(Optional.of(PostfachNachrichtTestFactory.SENT_AT)).when(service).getFirstSentAt(any());
			doReturn(answers).when(service).findAnswers(any(), any());
		}

		@Test
		void shouldCallFindAnswers() {
			enrichByAnswer();

			verify(service).findAnswers(SAML_TOKEN, RueckfrageTestFactory.ID);
		}

		@Test
		void shouldSetAnsweredAt() {
			var enrichedRueckfrage = enrichByAnswer();

			assertThat(enrichedRueckfrage.getAnsweredAt()).isEqualTo(PostfachNachrichtTestFactory.SENT_AT);
		}

		@Test
		void shouldSetStatus() {
			var enrichedRueckfrage = enrichByAnswer();

			assertThat(enrichedRueckfrage.getStatus()).isEqualTo(RueckfrageStatus.ANSWERED);
		}

		@Test
		void shouldReturnFullFilledValue() {
			var enrichedRueckfrage = enrichByAnswer();

			assertThat(enrichedRueckfrage).usingRecursiveComparison().ignoringFields("answeredAt", "vorgangId", "vorgangName", "status")
					.isEqualTo(rueckfrage);
		}

		private Rueckfrage enrichByAnswer() {
			return service.enrichByAnswers(rueckfrage, SAML_TOKEN);
		}
	}

	@DisplayName("Enrich rueckfrage")
	@Nested
	class TestEnrichRueckgabe {

		private final Rueckfrage rueckfrageWithoutVorgangName = RueckfrageTestFactory.createBuilder().vorgangName(null).build();
		private final List<Rueckfrage> rueckfragen = Collections.singletonList(rueckfrageWithoutVorgangName);

		@BeforeEach
		void mock() {
			doReturn(Stream.of(RueckfrageTestFactory.create())).when(service).addVorgangName(any(), any());
		}

		@Test
		void shouldCallAddVorgangNameIfRueckfragenExists() {
			service.enrichRueckfragen(rueckfragen);

			verify(service).addVorgangName(Collections.singletonList(rueckfrageWithoutVorgangName), RueckfrageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldReturnValue() {
			var enrichedRueckfragen = service.enrichRueckfragen(rueckfragen);

			assertThat(enrichedRueckfragen.findFirst()).get().extracting(Rueckfrage::getVorgangName).isEqualTo(RueckfrageTestFactory.VORGANG_NAME);
		}
	}

	@DisplayName("Add vorgang name")
	@Nested
	class TestAddVorgangName {

		private final Rueckfrage rueckfrage = RueckfrageTestFactory.createBuilder()
				.vorgangName(null)
				.build();

		@BeforeEach
		void mock() {
			when(vorgangService.getVorgang(any())).thenReturn(VorgangTestFactory.create());
		}

		@Test
		void shouldCallGetVorgang() {
			addVorgangName().toList();

			verify(vorgangService).getVorgang(RueckfrageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetVorgangName() {
			var rueckfragenWithVorgangName = addVorgangName().toList();

			assertThat(rueckfragenWithVorgangName.getFirst().getVorgangName()).isEqualTo(VorgangTestFactory.NAME);
		}

		private Stream<Rueckfrage> addVorgangName() {
			return service.addVorgangName(Collections.singletonList(rueckfrage), RueckfrageTestFactory.VORGANG_ID);
		}
	}

	@DisplayName("Get first sentAt")
	@Nested
	class TestGetFirstSentAt {
		private static final ZonedDateTime earlyZonedDateTime = ZonedDateTime.parse("2020-01-01T10:00:10Z");
		private static final PostfachNachricht earlyPostfachNachricht = PostfachNachrichtTestFactory.createBuilder().sentAt(earlyZonedDateTime)
				.build();

		private static final ZonedDateTime lateZonedDateTime = ZonedDateTime.parse("2021-01-01T10:00:10Z");
		private static final PostfachNachricht latePostfachNachricht = PostfachNachrichtTestFactory.createBuilder().sentAt(lateZonedDateTime).build();

		@Test
		void shouldReturnSentAt() {
			var firstSentAt = service.getFirstSentAt(Stream.of(PostfachNachrichtTestFactory.create(), earlyPostfachNachricht, latePostfachNachricht));

			assertThat(firstSentAt).contains(earlyZonedDateTime);
		}
	}

	@Nested
	class TestSendRueckfragenAnswers {
		static final String SAML_TOKEN = "TOKEN";
		static final String RUECKFRAGE_ID = UUID.randomUUID().toString();
		static final PostfachNachricht NACHRICHT = PostfachNachrichtTestFactory.create();

		@Test
		void shouldCallVerify() {
			service.sendRueckfrageAnswer(SAML_TOKEN, RUECKFRAGE_ID, NACHRICHT);

			verify(service).verifyToken(SAML_TOKEN);
		}

		@Test
		void shouldCallVerifier() {
			service.findRueckfragen(SAML_TOKEN);

			verify(verifier).verify(anyString());
		}

		@Test
		void shouldCallPostfachService() {
			service.sendRueckfrageAnswer(SAML_TOKEN, RUECKFRAGE_ID, NACHRICHT);

			verify(postfachNachrichtService).persistAnswer(eq(RUECKFRAGE_ID), any(PostfachNachricht.class));
		}
	}

	@Nested
	class TestFindAnswer {

		private static final String SAML_TOKEN = "TOKEN";

		@BeforeEach
		void init() {
			when(decrypter.decryptPostfachId(any())).thenReturn(GrpcFindRueckfrageRequestTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldCallVerify() {
			service.findAnswers(SAML_TOKEN, PostfachNachrichtTestFactory.ID);

			verify(service).verifyToken(SAML_TOKEN);
		}

		@Test
		void shouldCallFindAnswers() {
			service.findAnswers(SAML_TOKEN, PostfachNachrichtTestFactory.ID);

			verify(postfachNachrichtService).findAnswers(AntragraumService.BAYERN_ID_SERVICE_KONTO_TYPE,
					GrpcFindRueckfrageRequestTestFactory.POSTFACH_ID,
					PostfachNachrichtTestFactory.ID);
		}
	}

	@DisplayName("Get postfachId")
	@Nested
	class TestGetPostfachId {

		@Mock
		private Response response;

		private static final String SAML_TOKEN = "TOKEN";

		@BeforeEach
		void init() {
			when(parser.parse(anyString())).thenReturn(response);
			when(decrypter.decryptPostfachId(any())).thenReturn(GrpcFindRueckfrageRequestTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldCallParser() {
			service.getPostfachId(SAML_TOKEN);

			verify(parser).parse(SAML_TOKEN);
		}

		@Test
		void shouldCallDecrypter() {
			service.getPostfachId(SAML_TOKEN);

			verify(decrypter).decryptPostfachId(response);
		}
	}
}